/* elf.c -- elf support for cn.kuehne.reflection.d 

  version 2006-11-01 10:12:39

  Copyright (C) 2006 Thomas Kühne <thomas@kuehne.cn>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <elf.h>
#include <link.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>

void new_symbol(const char* name, void* value, size_t size, uint8_t type);

#define SymbolType_UNSPECIFIED	0
#define SymbolType_OBJECT	1
#define SymbolType_FUNCTION	2

#ifdef DEBUG
#define LOG printf
#define new_symbol(name, value, size, type) \
	{ printf("[%p\ts:%zu\tt:%hhu]\t%s\n", (value), (size), (type), (name)); } while(0);
#else
	void LOG(char* format, ...){
	}
#endif

#define SAVE_READ(buffer, count) if(read((fd), (buffer), (count)) < (ssize_t)(count)) return;
#define SAVE_SEEK(offset) if(lseek((fd), (offset), SEEK_SET) == -1) return;
static void scan_elf(void* start, void* end, int fd){
	ElfW(Ehdr) header;
	ElfW(Shdr) section;
	ElfW(Sym) sym;

	int i = 0;
	size_t j = 0;
	char* string_table = NULL;

	LOG("scan_elf(%p - %p)\n", start, end);

	/* read elf header */
	SAVE_READ(&header, sizeof(ElfW(Ehdr)));
	
	/* read section */
	if(header.e_shoff == 0){
		return;
	}
	
	for(i = header.e_shnum - 1; i > -1; i--){
		SAVE_SEEK(header.e_shoff + i * header.e_shentsize);
		SAVE_READ(&section, sizeof(ElfW(Shdr)));

		LOG("[%i] %i\n", i, section.sh_type);
		if(section.sh_type == SHT_STRTAB){
			/* read string table */
			LOG("[%i] is STRING\n", i);
			SAVE_SEEK(section.sh_offset);
			string_table = (char*) malloc(section.sh_size);
			SAVE_READ(string_table, section.sh_size);
		}else if(section.sh_type == SHT_SYMTAB){
			/* read symtab */
			LOG("[%i] is SYMTAB\n", i);
			if(section.sh_offset){
				for(j = 0; j < section.sh_size / sizeof(ElfW(Sym)); j++){
					SAVE_SEEK(section.sh_offset + j * sizeof(ElfW(Sym)));
					SAVE_READ(&sym, sizeof(ElfW(Sym)));
					if(!sym.st_name || !sym.st_value){
						continue;
					}
					/* >> KLUDGE : odd st_info */
					sym.st_info &= 0xF;
					/* << KLUDGE : odd st_info */
					switch(sym.st_info){
						case STT_NOTYPE:
							sym.st_info = SymbolType_UNSPECIFIED;
							break;
						case STT_OBJECT:
							sym.st_info = SymbolType_OBJECT;
							break;
						case STT_FUNC:
							sym.st_info = SymbolType_FUNCTION;
							break;
						default:
							continue;
					}
					/* >> KLUDGE : address */
					if(sym.st_value < (size_t)start){
						sym.st_value += (size_t)start;
					}
					/* << KLUDGE : address */

					new_symbol(string_table + sym.st_name, (void*)sym.st_value, sym.st_size, sym.st_info);
				}
			}
			return;
		}
	}
}
#undef SAVE_READ
#undef SAVE_SEEK

void scan_binary(){
	char* line;
	size_t start;
	size_t end;
	FILE* map;
	int i;
	char tmp;
	char* buffer;
	size_t buffer_len = 4096 * 2;

	LOG("scan_binary()\n");

	buffer = (char*) malloc(buffer_len);

	map = fopen("/proc/self/maps", "r");
	if(map == NULL){
		return;
	}
	
	while((line = fgets(buffer, buffer_len, map)) != NULL){
		LOG("line: %s\n", line);
		i = sscanf(line, "%zx-%zx r-x%*c %*x %*c%*c:%*c%*c %*u%c",
			&start, &end, &tmp);
		if((i != 3) || (tmp != ' ') || (line[73] != '/')){
			continue;
		}
	
		if(memcmp((void*)start, "\177ELF", 4)){
			continue;
#ifdef ELFCLASS32
		}else if(sizeof(size_t) == 4){
			if(((char*)start)[EI_CLASS] != ELFCLASS32){
				continue;
			}
#endif
#ifdef ELFCLASS64
		}else if(sizeof(size_t) == 8){
			if(((char*)start)[EI_CLASS] != ELFCLASS64){
				continue;
			}
#endif
		}else{
			continue;
		}

		line+= 73;
		line[strlen(line)-1] = 0;
		i = open(line, O_RDONLY);
		LOG("open(%s) -> %i\n", line, i);
		if(i == -1){
			continue;
		}

		scan_elf((void*)start, (void*)end, i);
		close(i);
	}

	fclose(map);
	free(buffer);
}

#ifdef DEBUG
int main(){
	printf("PID		%i\n", getpid());
	printf("scan_elf	%p\n", &scan_elf);
	printf("scan_binary	%p\n", &scan_binary);
	printf("main	 	%p\n", &main);
	scan_binary();
	getchar();

	return 0;
}
#endif
