/* reflection.d -- stub of reflection support for the D programming language

  version 2006-11-01 10:12:39

  Copyright (C) 2006 Thomas Kühne <thomas@kuehne.cn>

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
*/


module cn.kuehne.reflection;

private import std.string;
private import std.regexp;

version(linux){
	// DMD <-> GDC
	version = Unix;
}

extern(C) void scan_binary();

extern(C){
	void new_symbol(char* name, void* value, size_t size, SymbolType type){
		Symbol s = new Symbol(name, value, size, type);
		symbols[s.mangled_name] = s;
	}
}

static this(){
	scan_binary();
}

Symbol[char[]] symbols;

enum SymbolType : ubyte{
	UNSPECIFIED	= 0,
	OBJECT		= 1,
	FUNCTION	= 2
}

enum CallingConvention{
	C,
	Pascal,

	// aliases
	stdcall = Pascal
}


struct Argument{
	size_t size;
	void* data;
}

struct Return{
	int	_int;
	long	_long;
	union{
		void* 	pointer;
		size_t	_size_t;
	}
	real	_real;
}

class ReflectionException : Exception{
	this(char[] msg){
		super(msg);
	}
}

class SymbolException : ReflectionException{
	this(char[] msg){
		super(msg);
	}
}

class Symbol{
	char[] mangled_name;
	SymbolType type;
	void* start;
	size_t size;

	this(char* name, void* start, size_t size, SymbolType type){
		mangled_name = std.string.toString(name);
		this.start = start;
		this.size = size;
		this.type = type;
	}

	static void filter(inout Symbol[char[]] sym, char[] pattern, bool remove_match = false){
		filter(sym, new RegExp(pattern), remove_match);
	}

	static void filter(inout Symbol[char[]] sym, RegExp pattern, bool remove_match = false){
		if(remove_match){
			foreach(name; sym.keys.dup){
				if(pattern.match(name)){
					sym.remove(name);
				}
			}
		}else{
			foreach(name; sym.keys.dup){
				if(!pattern.match(name)){
					sym.remove(name);
				}
			}
		}
	}

	Return call(CallingConvention conv, Argument[] args){
		return call(conv, start, args);
	}

	static Return call(CallingConvention conv, void* func, Argument[] args){
		switch(conv){
			case CallingConvention.C:{
				return callC(func, args);
			}case CallingConvention.Pascal:{
				return callPascal(func, args);
			}default:{
				throw new SymbolException("unknown calling convention");
			}
		}
	}

	static Return callC(void* func, Argument[] arguments){
		return callClike(func, arguments, false);
	}

	private static Return callClike(void* func, Argument[] arguments, bool isPascal){
		version(Unix){
			static if(size_t.sizeof == 4){
				int _EAX_;
				int _EDX_;
				real _ST0_;
				size_t* size;
				size_t* arg;
				size_t cleanup;
				size = new size_t[arguments.length + 1];
				arg = new size_t[arguments.length + 1];

				foreach(index, argument; isPascal ? arguments : arguments.dup.reverse){
					size[index] = argument.size;
					arg[index] = cast(size_t) argument.data;
				}
				size[arguments.length] = 0; // end marker
				arg[arguments.length] = 0; // end marker

				/*
				 * don't change:
				 *	EBX, ESI, EDI, EBP
				 *	DS, ES, SS
				 */
				asm{
				next_argument:
					// get argument size
					mov EAX, size;
					add size, 4;
					mov EDX, [EAX];
					// last argument ?
					cmp EDX, 0;
					je do_call;
					// get data
					mov EAX, arg;
					add arg, 4;
					mov ECX, [EAX];
					// check argument size 
					cmp EDX, 4;
					jbe still_more;
					// argument is bigger than 32bits
					mov EAX, EDX;
					and EAX, 0xFF_FF_FF_F8;
					sub EAX, 4;
					add ECX, EAX;
				still_more:
					cmp EDX, 1;
					jl next_argument;
					mov EAX, [ECX]; // BUG: movzx for 1-3byte arguments?
					push EAX;
					sub ECX, 4;
					add cleanup, 4;
					sub EDX, 4;
					jmp still_more;
				do_call:
					call func;
					// adjust stack
					cmp isPascal, 1;
					je no_clean;
					add ESP, cleanup;
				no_clean:
					// store results
					mov _EAX_, EAX;
					mov _EDX_, EDX;
					fstp _ST0_;
					fld _ST0_;
				}

				Return result;
				result._size_t = _EAX_;
				result._int = _EAX_;
				result._real = _ST0_;
				result._long = cast(uint) _EDX_;
				result._long <<= 32;
				result._long |= cast(uint) _EAX_;
				return result;
			}else{
				static assert(0, "unimplemented pointer size");
			}
		}else{
			static assert(0, "unimplemented system");
		}
	}

	static Return callPascal(void* func, Argument[] arguments){
		return callClike(func, arguments, true);
	}

	char[] toString(){
		return format("%s [start:%8s size:%4s type:%s]", mangled_name, start, size, .toString(type));
	}
}

Symbol findSymbol(void* addr){
	foreach(s; symbols){
		if((s.start <= addr) && (addr < s.start + s.size)){
			return s;
		}
	}
	return null;
}

char[] addressToString(void* addr){
	Symbol s = findSymbol(addr);
	if(s){
		if(s.start == addr){
			return s.mangled_name;
		}else{
			return format("%s+0x%X", s.mangled_name, addr - cast(size_t)s.start);
		}
	}
	return format("0x%08X", addr);
}

char[] toString(SymbolType t){
	const char[][] v = ["UNSPECIFIED", "OBJECT", "FUNCTION"];
	return v[t];
}

CallingConvention toCallingConvention(char[] name){
	switch(name){
		case "C", "cdecl", "_cdecl", "__cdecl":
			return CallingConvention.C;
		case "Pascal", "stdcall", "_stdcall", "__stdcall":
			return CallingConvention.Pascal;
		default:
			throw new Exception("unsupported calling convention '"~ name ~"'");
	}
}

version(Demo){
	import std.stdio;

	extern(C) static void* addr_c;
	extern(Pascal) static void* addr_p;
	static void* addr_d;

	extern(Pascal) long pascal_sum(byte a, long l, short s){
		l += a;
		l += s;
		return l;
	}

	void i_am_at(){
		writefln("hello, my name is: %s", addressToString(&i_am_at));
		writefln("\tthis static C variable is called: %s", addressToString(&addr_c));
		writefln("\tthis static D variable is called: %s", addressToString(&addr_d));
		writefln("\tthis static Pascal variable is called: %s", addressToString(&addr_p));

		void* addr;
		writefln("\tthis stack variable is called: %s", addressToString(&addr));
		writefln();
	}

	void all_std_uni_symbols(){
		writefln("the symbols defined by std.uni are:");

		foreach(s; symbols){
			if(std.string.find(s.mangled_name, "3std3uni") >= 0){
				writefln("\t%s", s);
			}
		}
		writefln();
	}

	void call_a_c_function(){
		char[] funct = "strlen";
		writefln("let's call the C function %s", funct);
		Symbol* s = funct in symbols;
		if(s){
			Argument[] arguments = new Argument[1];
			char* string = "See the summary chart at the beginning";
			arguments[0].size = string.sizeof;
			arguments[0].data = &string;

			Return result = s.call(CallingConvention.C, arguments);
			writefln("\t%s(\"%s\") -> %s",  funct, std.string.toString(string), result._size_t);
		}else{
			writefln("\tsorry, there seems to be no \"%s\"", funct);	
		}
		writefln();
	}

	void call_a_pascal_function(){
		char[] funct = "PASCAL_SUM";
		writefln("let's call the pascal function %s", funct);
		Symbol* symbol = funct in symbols;
		if(symbol){
			Argument[] arguments = new Argument[3];
			byte b = 0x44;
			arguments[0].size = b.sizeof;
			arguments[0].data = &b;
			long l = 0xAABB_5566_7788_9900L;
			arguments[1].size = l.sizeof;
			arguments[1].data = &l;
			short s = 0x11_22;
			arguments[2].size = s.sizeof;
			arguments[2].data = &s;

			Return result = symbol.call(CallingConvention.Pascal, arguments);
			writefln("\t%s(0x%X, 0x%X, 0x%X) -> 0x%X", funct, b, l, s, result._long);
		}else{
			writefln("\tsorry, there seems to be no \"%s\"", funct);	
		}
		writefln();
	}


	int main(){
		i_am_at();
		all_std_uni_symbols();
		call_a_c_function();
		call_a_pascal_function();

		return 0;
	}
}
